package com.urbandroid.sleep

import android.app.AlarmManager
import android.app.PendingIntent
import android.app.Service
import android.content.Context
import android.content.Context.ALARM_SERVICE
import android.content.Intent
import android.content.SharedPreferences
import android.hardware.Sensor
import android.hardware.SensorManager
import android.os.Build
import android.preference.PreferenceManager
import android.support.annotation.DrawableRes
import android.support.annotation.RequiresApi
import android.support.v4.app.NotificationCompat
import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.service.Settings
import com.urbandroid.util.Experiments
import com.urbandroid.util.SleepPermissionCompat

/**
 * Send explicit broadcast from Android O (Api level >= 26) by setting package name into intent.
 * If no package is specified context.package is used
 * If intent has already configured either packageName or componentName - no change will be done to intent
 * If more then one package name is specified - separate broadcast is sent per package
 */
@JvmOverloads
fun Context.sendExplicitBroadcast(intent: Intent, vararg packages: String = arrayOf(packageName)) {
    if (Experiments.getInstance()?.isBackgroundLimitationApi26 == true && intent.hasNoPackage()) {
        packages.forEach { sendBroadcast(intent.copy(it)) }
        return
    }
    sendBroadcast(intent)
}

fun Context.getAlarmManager(): AlarmManager = getSystemService(ALARM_SERVICE) as AlarmManager

val Context.settings: Settings
    get() = Settings(this)

val Context.preferences: SharedPreferences
    get() = PreferenceManager.getDefaultSharedPreferences(this)

/**
 * Copy intent with new given package
 */
fun Intent.copy(pkg: String) = Intent(this).apply { `package` = pkg }
fun Intent.hasNoPackage() = `package` == null && component == null
fun Intent.hasPackage() = !hasNoPackage()

/**
 * StopSelf preceded by startForeground call - needed for Android O background limitation.
 * It should be used instead of call Service.stopSelf() in case where service is foreground service and
 * startForeground method was not called yet.
 */
@JvmOverloads
fun Service.stopForegroundSelf(channelId: String, notificationId: Int, @DrawableRes smallIcon: Int = R.drawable.empty, forceStartForeground: Boolean = true) {
    if (forceStartForeground) {
        try {
            val notificationBuilder = NotificationCompat.Builder(this, channelId).apply {
                setContentIntent(PendingIntent.getActivity(applicationContext, 0, Intent(), 0))
                color = resources.getColor(R.color.tint_dark)
                setContentTitle(getString(R.string.good_night))
                setContentText("")
                setSmallIcon(smallIcon)
            }


            startForeground(notificationId, notificationBuilder.build())
        }catch (ex: Exception) {
            if (Logger.isInitialized()) {
                Logger.logSevere("startForeground failure", ex)
            }
        }
    }
    stopSelf()
}

fun Context.isPermissionGranted(vararg permissions: String): Boolean {
    permissions.forEach {
        if (!SleepPermissionCompat.isPermissionGranted(this, it)) {
            return false
        }
    }
    return true
}

fun Context.targetSdkVersion() = packageManager.getPackageInfo(packageName, 0).applicationInfo.targetSdkVersion

@RequiresApi(Build.VERSION_CODES.KITKAT)
fun Context.accelSensorBatchingQueueSize(): Int {
    val sensorManager = getSystemService(Context.SENSOR_SERVICE) as SensorManager
    val accelSensor = sensorManager.getDefaultSensor(Sensor.TYPE_ACCELEROMETER)
    return accelSensor?.fifoMaxEventCount?:0
}

fun Context.isSensorBatchingSupported(): Boolean = Settings.isSensorBatchingSupported(this)

@JvmField
val PACKAGES_3RD_PARTY = arrayOf(
    "com.urbandroid.sleep.addon.generic.samsung",
    "com.urbandroid.sleep.garmin",
    "cz.zdenekhorak.mibandtools",
    "cz.zdenekhorak.amazfittools",
    "com.mc.miband1"
)

@JvmField
val PACKAGE_PEBBLE = "com.getpebble.android.basalt"
@JvmField
val PACKAGE_TASKER = "net.dinglisch.android.taskerm"