package com.urbandroid.sleep.activityrecognition

import android.content.Context
import com.google.android.gms.location.ActivityTransitionEvent
import com.google.android.gms.location.DetectedActivity
import com.urbandroid.common.logging.Logger
import com.urbandroid.util.Experiments
import java.io.*
import java.lang.StringBuilder
import java.text.SimpleDateFormat
import java.util.*
import java.util.concurrent.TimeUnit
import java.util.concurrent.TimeUnit.HOURS
import java.util.concurrent.TimeUnit.MILLISECONDS

class ActivityIntervals(private val keepHours: Long = 48, var currentTime: (() -> Long)) {

    val intervals = mutableListOf<Interval>()

    constructor() : this(currentTime = {System.currentTimeMillis()})

    fun from(input: ByteArray) = from(ByteArrayInputStream(input))
    fun from(intervals: List<Interval>): ActivityIntervals = this.intervals.let { it.addAll(intervals); this }

    fun from(input: InputStream): ActivityIntervals {
        intervals.clear()
        val minTime = currentTime() - HOURS.toMillis(keepHours)
        try {
            val oos = ObjectInputStream(input)
            while (true) {
                val from = oos.readLong()
                val to = oos.readLong()
                if (from < minTime && to < minTime) {
                    // intervals older than declared keepHours are deleted/ignored
                    continue
                }
                intervals.add(Interval(from, to))
            }
        } catch (eofe: EOFException) {
        } finally {
            try { input.close() } catch (ignore: Exception) {}
        }
        return this
    }

    fun shift(value:Long, unit: TimeUnit): ActivityIntervals {
        val shiftedIntervals = mutableListOf<Interval>()
        intervals.forEach {
            shiftedIntervals.add(it.shift(value, unit))
        }
        intervals.clear()
        intervals.addAll(shiftedIntervals)
        return this
    }

    fun size() = intervals.size
    fun isEmpty() = intervals.isEmpty()

    @Throws(IOException::class)
    fun export(file: File) = export(file.outputStream())

    fun exportToBytes(): ByteArray {
        val bos = ByteArrayOutputStream(BUFFER_SIZE)
        export(bos)
        return bos.toByteArray()
    }

    @Throws(IOException::class)
    fun export(os: OutputStream) = ObjectOutputStream(os).use {
            oos ->
                try {
                    intervals.forEach { oos.apply { writeLong(it.from); writeLong(it.to) } }
                }catch (ex: Exception) {
                    Logger.logWarning("ActivityIntervals: export failure", ex)
                }
            this
        }

    fun add(interval: Interval): ActivityIntervals {
        intervals.add(interval)
        return this
    }

    fun add(transition: ActivityTransitionEvent): ActivityIntervals {
        val activity = if (transition.isEnter) DetectedActivity(DetectedActivity.STILL, 100) else DetectedActivity(DetectedActivity.TILTING, 100)
        return add(activity, transition.timestamp)
    }

    fun add(activity: DetectedActivity, time: Long = currentTime()): ActivityIntervals {

        val isStill = activity.type == DetectedActivity.STILL
        val isAtLeast75 = activity.confidence >= MINIMAL_CONFIDENCE
        if (!isAtLeast75) {
            //Logger.logDebug("Discarding $activity since it is not at least $MINIMAL_CONFIDENCE %")
            return this
        }

        val interval = intervals.firstOrNull { time in it }
        if (interval == null) {
            if (!isStill) {
                if (Experiments.getInstance().isOurExperimentalPhone) {
                    Logger.logDebug("Discarding $activity since no interval found containing ${format(time)} \n ${toString(true)}")
                }
                return this
            }
            if (intervals.isEmpty() || (intervals.last().isClosed && intervals.last().to < time)) {
                intervals.add(Interval(time))
                return this
            }

            // first nearby interval which is after 'time'
            val firstAfter = intervals.firstOrNull { it.from > time }
            if (intervals.isNotEmpty() && firstAfter != null) {
                with(intervals) {
                    val extended = firstAfter.copy(from = time)
                    val index = intervals.indices.firstOrNull { intervals[it] == firstAfter }
                    if (index != null) {
                        removeAt(index)
                        add(index, extended)
                    }
                }
            }
            return this
        }

        if (interval.isOpen) {
            if (!isStill) {
                intervals.remove(interval)
                intervals.add(interval.copy(to = time))
            }
            return this
        }

        // is closed
        if (isStill) {
            if (intervals.last().to < time) {
                intervals.add(Interval(time))
            }
        } else {
            val index = intervals.indexOfFirst { it == interval }
            if (index != -1) {
                val pair = interval.splitBy(time)
                if (pair != null) {
                    with(intervals) {
                        removeAt(index)
                        add(index, pair.first)
                        add(index + 1, pair.second)
                    }
                }
            }
        }
        return this
    }

    fun toString(userFriendly: Boolean): String {
        var previous: Long = -1
        val result = StringBuilder()
        for (interval in intervals) {
            if (userFriendly) {
                result.append("[${format(interval.from, previous)},${format(interval.to, interval.from)}]")
                previous = interval.to
            } else {
                result.append("[${interval.from},${interval.to}]")
            }
        }
        return result.toString()
    }

    data class Interval @JvmOverloads constructor(val from: Long, val to: Long = -1) {
        val isOpen: Boolean
            get() = to == -1L

        val isClosed: Boolean
            get() = !isOpen

        fun length(unit: TimeUnit = HOURS): Long {
            return unit.convert(to - from, MILLISECONDS)
        }

        fun intersection(other: Interval): Interval? = when {
            from >= other.from && to <= other.to -> this
            other.from >= from && other.to <= to -> other
            from >= other.from && from <= other.to -> Interval(from, other.to)
            other.from in from..to -> Interval(other.from, to)
            else -> null
        }

        fun shift(value: Long, unit: TimeUnit) = Interval(
                from + MILLISECONDS.convert(value, unit),
                if (to == -1L) -1L else to + MILLISECONDS.convert(value, unit)
        )
        operator fun contains(time: Long) = if (to == -1L) time >= from else time in from..to

        fun splitBy(time: Long): Pair<Interval, Interval>? {
            if (time !in this) {
                return null
            }
            return Interval(from, time) to Interval(time+1, to)
        }

        override fun toString() = "[${format(from, -1)},${format(to, from)}]"
    }

    companion object {

        private const val EXPORT_FILE = "activity-recognition.dat"
        private const val TIME_FORMAT = "HH:mm"
        const val FULL_FORMAT = "yyyy-MM-dd HH:mm"
        const val DAY_TIME_FORMAT = "EEE HH:mm"
        const val DAY_TIME_SECOND_FORMAT = "EEE HH:mm:ss"
        const val MINIMAL_CONFIDENCE = 75
        const val BUFFER_SIZE = 8192

        fun getStorageFile(context: Context) = File(context.cacheDir, EXPORT_FILE)

        fun from(intervals: List<Interval>): ActivityIntervals = ActivityIntervals().from(intervals)
        fun from(input: InputStream): ActivityIntervals = ActivityIntervals().from(input)
        fun from(input: ByteArray): ActivityIntervals = ActivityIntervals().from(ByteArrayInputStream(input))
        fun from(input: File): ActivityIntervals =
                if (input.exists()) {
                    val stream = input.inputStream()
                    try {
                        stream.use { from(stream) }
                    } catch (e: IOException) {
                        Logger.logDebug("corrupted file: ${input.absolutePath} $e")
                        // somehow file is corrupted - so delete it
                        try { input.delete() } catch (ignore: Exception) {}
                        ActivityIntervals()
                    }
                } else {
                    Logger.logDebug("not found file: ${input.absolutePath}")
                    ActivityIntervals()
                }

        fun formatWithDay(date: Long) = format(date, date - TimeUnit.DAYS.toMillis(2))

        fun format(date: Long, previous: Long = -1): String {
            if (date == -1L) {
                return "-"
            }
            val cal = Calendar.getInstance().apply { timeInMillis = date }
            val calPrevious = Calendar.getInstance().apply { timeInMillis = previous }
            val dateFormat = if (previous != -1L && cal.get(Calendar.DAY_OF_YEAR) == calPrevious.get(Calendar.DAY_OF_YEAR)) TIME_FORMAT else DAY_TIME_FORMAT
            return SimpleDateFormat(dateFormat).format(cal.time)
        }


    }

}
