package com.urbandroid.sleep.activityrecognition

import android.app.PendingIntent
import android.content.Context
import android.content.Intent
import android.os.Build
import com.google.android.gms.location.ActivityRecognition
import com.google.android.gms.location.ActivityTransition.*
import com.google.android.gms.location.ActivityTransitionRequest
import com.google.android.gms.location.DetectedActivity
import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.PendingIntentBuilder
import com.urbandroid.sleep.accelSensorBatchingQueueSize
import com.urbandroid.sleep.activityrecognition.SleepTimeSuggestionManager.Mode.DISABLED
import com.urbandroid.sleep.settings
import com.urbandroid.util.Experiments

class ActivityRecognitionInitializer(private val context: Context) {

    fun initialize() {
        val experiments = Experiments.getInstance()
        if (!experiments.isActivityRecognitionExperiment) {
            return
        }
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {

            val batchingQueueSize = context.accelSensorBatchingQueueSize()
            if (batchingQueueSize < MIN_BATCHING_QUEUE_SIZE) {
                Logger.logWarning("ActivityRecognition disabled - too small or no sensor batching: $batchingQueueSize")
                context.settings.sleepTimeSuggestionMode = DISABLED
                return
            }

            val activityRecognitionClient = ActivityRecognition.getClient(context)

            val transitions = arrayOf(
                    Builder().apply {
                        setActivityType(DetectedActivity.STILL)
                        setActivityTransition(ACTIVITY_TRANSITION_ENTER)
                    }.build(),
                    Builder()
                            .setActivityType(DetectedActivity.STILL)
                            .setActivityTransition(ACTIVITY_TRANSITION_EXIT)
                            .build())

            val request = ActivityTransitionRequest(transitions.toMutableList())
            activityRecognitionClient
                    .requestActivityTransitionUpdates(request, getActivityTransitionPendingIntent())
                    .addOnSuccessListener { Logger.logInfo("activity_transition_updates_successfully_enabled") }
                    .addOnFailureListener { Logger.logInfo("activity_transition_updates_not_enabled") }

            SleepTimeSuggestionManager(context).schedule()
            Logger.logInfo("ActivityRecognition initialized (batching: $batchingQueueSize)")
        }
    }

    fun destroy() {
        if (!Experiments.getInstance().isActivityRecognitionExperiment) {
            return
        }
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            val activityRecognitionClient = ActivityRecognition.getClient(context)
            activityRecognitionClient.removeActivityTransitionUpdates(getActivityTransitionPendingIntent())
            SleepTimeSuggestionManager(context).cancelAlarm()
            Logger.logInfo("destroyed")
        }
    }

    fun destroyOldActivityDetection() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            try {
                val activityRecognitionClient = ActivityRecognition.getClient(context)
                activityRecognitionClient.removeActivityUpdates(getActivityDetectionPendingIntent())
                SleepTimeSuggestionManager(context).cancelAlarm()
                Logger.logInfo("destroyOldActivityDetection")
            } catch (ex: Exception) {
                Logger.logSevere("destroyOldActivityDetection", ex)
            }
        }
    }
    private fun getActivityDetectionPendingIntent(): PendingIntent? {
        return PendingIntentBuilder.get(context, 0, Intent(context, ActivityReceiver::class.java), PendingIntent.FLAG_UPDATE_CURRENT).getExplicitBroadcast()
    }

    private fun getActivityTransitionPendingIntent(): PendingIntent? {
        return PendingIntentBuilder.get(context, 0, Intent(context, ActivityTransitionReceiver::class.java), PendingIntent.FLAG_UPDATE_CURRENT).getExplicitBroadcast()
    }

    companion object {
        const val MIN_BATCHING_QUEUE_SIZE = 500L

        fun from(context: Context) = ActivityRecognitionInitializer(context)
    }
}
