package com.urbandroid.sleep.activityrecognition

import android.annotation.SuppressLint
import android.app.AlarmManager
import android.app.PendingIntent
import android.content.Context
import android.content.Intent
import android.content.Intent.FLAG_ACTIVITY_NEW_TASK
import android.os.Parcelable
import android.support.v4.app.NotificationCompat
import android.support.v4.app.NotificationManagerCompat
import com.urbandroid.common.dayOfYear
import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.*
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeCalculator.Companion.from
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeCalculator.Companion.to
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeCalculator.Estimate
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeCalculator.Estimate.Type.ACTIVITY
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeCalculator.Estimate.Type.BASIC
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeSuggestion
import com.urbandroid.sleep.alarmclock.settings.TrackSettingsActivity
import com.urbandroid.sleep.alarmclock.settings.TrackSettingsActivity.EXTRA_SHOW_SLEEP_SUGGESTION_MODES
import com.urbandroid.sleep.domain.DuplicateRecordException
import com.urbandroid.sleep.domain.SleepRecord
import com.urbandroid.sleep.notification.NOTIFICATION_CHANNEL_ID_SLEEP_TIME_SUGGESTION
import com.urbandroid.sleep.service.SharedApplicationContext
import com.urbandroid.sleep.service.sync.RequestSyncReceiver
import java.util.*

class SleepTimeSuggestionManager(private val context: Context) {

    enum class Mode(val code: Int) {
        DISABLED(0), MANUAL(1), AUTOMATIC(2);

        companion object {
            @JvmStatic
            fun findOrNull(code: Int) = values().firstOrNull { it.code == code }
        }
    }

    @SuppressLint("StringFormatInvalid")
    fun suggest(mode: Mode) {
        if (mode == SleepTimeSuggestionManager.Mode.DISABLED) return

        val estimate = SleepTimeSuggestion().calculate(context, Estimate(BASIC, from(), to()))

        val record: SleepRecord? = if (mode == Mode.AUTOMATIC) {
            if (estimate.type != ACTIVITY) {
                Logger.logInfo("Automatic mode is enabled but ${estimate.type} was not calculated from detected activities")
                return
            }
            // save estimate recod to db
            addRecord(estimate)
        } else { null }

        val actionSettingsPendingIntent = PendingIntent.getActivity(
                context, NOTIFICATION_ID,
                Intent(context, TrackSettingsActivity::class.java).apply {
                    addFlags(FLAG_ACTIVITY_NEW_TASK)
                    putExtra(NOTIFICATION_ID_NAME, NOTIFICATION_ID)
                    putExtra(EXTRA_SHOW_SLEEP_SUGGESTION_MODES, true)
                },
                PendingIntent.FLAG_UPDATE_CURRENT)

        val estimateText = context.getString(
                R.string.sleep_time_suggestion_notification,
                ActivityIntervals.formatWithDay(estimate.from.timeInMillis),
                ActivityIntervals.format(estimate.to.timeInMillis, estimate.from.timeInMillis))

        val notificationBuilder = NotificationCompat.Builder(context, NOTIFICATION_CHANNEL_ID_SLEEP_TIME_SUGGESTION)
                .setColor(context.resources.getColor(R.color.tint_dark))
                .setAutoCancel(true)
                .setContentText(context.getString(R.string.sleep_time_suggestion_description))
                .setContentTitle(estimateText)
                .setSmallIcon(R.drawable.ic_action_track_white)

        with(notificationBuilder) {
            if (context.settings.sleepTimeSuggestionMode == Mode.AUTOMATIC) {
                val actionEditRecordPendingIntent =
                    PendingIntent.getActivity(
                            context, NOTIFICATION_ID,
                            Intent(context, GraphDetailsActivity::class.java).apply {
                                addFlags(FLAG_ACTIVITY_NEW_TASK)
                                putExtra(SleepRecord.PARCELABLE_PARAM_NAME, record as Parcelable)
                            },
                            PendingIntent.FLAG_UPDATE_CURRENT)

                addAction(0, context.getString(R.string.edit_record), actionEditRecordPendingIntent)
                setContentIntent(actionEditRecordPendingIntent)
            } else {
                val actionAddRecordPendingIntent = PendingIntent.getActivity(
                        context, NOTIFICATION_ID,
                        Intent(context, NewAddRecordActivity::class.java).apply {
                            addFlags(FLAG_ACTIVITY_NEW_TASK)
                            putExtra(EXTRA_SLEEP_SUGGESTION, true)
                            putExtra(NewAddRecordActivity.RANGE_FROM, from().timeInMillis)
                            putExtra(NewAddRecordActivity.RANGE_TO, to().timeInMillis)
                        },
                        PendingIntent.FLAG_UPDATE_CURRENT)

                addAction(0, context.getString(R.string.add_record_button), actionAddRecordPendingIntent)
                setContentIntent(actionAddRecordPendingIntent)
            }
            addAction(0, context.getString(R.string.settings), actionSettingsPendingIntent)
        }

        NotificationManagerCompat.from(context).notify(NOTIFICATION_ID, notificationBuilder.build())

    }

    fun schedule(){
        // at 3pm
        val calendar = Calendar.getInstance().apply {
            if (get(Calendar.HOUR_OF_DAY) >= 15) {
                add(Calendar.DAY_OF_YEAR, 1)
            }
            set(Calendar.HOUR_OF_DAY, 15)
            set(Calendar.MINUTE, 0)
            add(Calendar.SECOND, 0)
            add(Calendar.MILLISECOND, 0)
            add(Calendar.MINUTE, 1)
        }

        val settings = context.settings
        if (settings.sleepTimeSuggestionLastDay == calendar.dayOfYear) {
            Logger.logDebug("Notify Sleep Time Suggestion already scheduled for ${calendar.time} day of year is: ${calendar.dayOfYear}")
            return
        }
        val pendingIntent = prepareIntent()
        settings.sleepTimeSuggestionLastDay = calendar.dayOfYear
        context.getAlarmManager().apply { set(AlarmManager.RTC, calendar.timeInMillis, pendingIntent.getExplicitBroadcast()) }

        Logger.logDebug("Notify Sleep Time Suggestion scheduled ${calendar.time}")
    }

    fun cancelAlarm(){
        prepareIntent().cancelAlarmService()
        context.settings.sleepTimeSuggestionLastDay = -1
        Logger.logDebug("Cancel Sleep Time Suggestion alarm")
    }

    fun cancel() {
        NotificationManagerCompat.from(context).cancel(NOTIFICATION_ID)
    }

    private fun addRecord(estimate: Estimate): SleepRecord? {
        val newRecord = SleepRecord(TimeZone.getDefault().id, estimate.from.time, estimate.to.time)
        newRecord.updateLatestTo(estimate.to.time)
        newRecord.to = estimate.to.time
        newRecord.isFinished = true
        newRecord.rateAndComment("#auto", 0f)

        // TODO find overlap

        newRecord.setHistory(listOf(0f))
        try {
            SharedApplicationContext.getInstance().sleepRecordRepository.addNewSleepRecord(newRecord)
            context.settings.incRecordsCount()

            SharedApplicationContext.getInstance().firebaseAnalyticsManager.setEventAutomaticRecordAddedFromSuggestion()

            Logger.logInfo("SYNC data changed broadcast")
            context.sendExplicitBroadcast(Intent(RequestSyncReceiver.REQUEST_SYNC_ACTION))
            Logger.logInfo("Automatic save $newRecord")
        } catch (e: DuplicateRecordException) {
            Logger.logWarning("Automatic add record", e)
        }

        return newRecord
    }


    private fun prepareIntent(): PendingIntentBuilder {
        val intent = Intent(context, NotifySleepTimeSuggestionReceiver::class.java)
        return PendingIntentBuilder(context, 0, intent, PendingIntent.FLAG_UPDATE_CURRENT)
    }

    companion object {
        const val NOTIFICATION_ID = 2442
        const val NOTIFICATION_ID_NAME = "notification_id"
        const val EXTRA_SLEEP_SUGGESTION = "sleep_suggestion"
    }
}