package com.urbandroid.sleep.activityrecognition.calculator

import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeCalculator.Companion.DAY
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeCalculator.Estimate
import com.urbandroid.sleep.domain.SleepRecord
import com.urbandroid.sleep.domain.SleepRecords
import com.urbandroid.sleep.service.SharedApplicationContext
import java.util.*
import java.util.Calendar.*

class SleepTimeDbRecordCalculator @JvmOverloads constructor(
        private val ignoreOverlapping: Boolean = false,
        private val recordCount: Int = 200): SleepTimeCalculator {

    /**
     * Estimate/Calculate start.end sleep time by finding average start/end for given week day
     */
    override fun estimate(estimate: Estimate): Estimate {

        Logger.logInfo("DB start range: $estimate")

        val fromCalendar = estimate.from.clone() as Calendar
        val toCalendar = estimate.to.clone() as Calendar

        val records = SharedApplicationContext.getInstance().sleepRecordRepository.getSleepRecords(0, recordCount, false)

        val cal = getInstance()

        var durationSum = 0
        var wakeupHourSum = 0
        var count = 0

        for (record in records) {
            cal.time = record.to
            if (record.sleepLength != 0 && record.sleepLength > 240 &&
                    cal.get(DAY_OF_WEEK) == toCalendar.get(DAY_OF_WEEK) &&
                    cal.get(HOUR_OF_DAY) <= 12) {
                durationSum += record.sleepLength
                wakeupHourSum += cal.get(MINUTE) + cal.get(HOUR_OF_DAY) * 60
                count++
            }
        }

        Logger.logInfo("Count $count")
        var estimateType = Estimate.Type.BASIC

        if (count > 2) {
            estimateType = Estimate.Type.DB
            val durationAvg = durationSum / count
            val wakeUpHourAvg = wakeupHourSum / count

            Logger.logInfo("AVG duration $durationAvg")
            Logger.logInfo("AVG wake up $wakeUpHourAvg")

            toCalendar.set(HOUR_OF_DAY, wakeUpHourAvg / 60)
            toCalendar.set(MINUTE, wakeUpHourAvg % 60)

            if (toCalendar.timeInMillis > System.currentTimeMillis() && toCalendar.timeInMillis - System.currentTimeMillis() < 240000) {
                toCalendar.timeInMillis = System.currentTimeMillis()
            }

            fromCalendar.time = toCalendar.time
            fromCalendar.add(MINUTE, -durationAvg)

            // round
            cal.timeInMillis = roundTime(fromCalendar.timeInMillis)
            fromCalendar.set(HOUR_OF_DAY, cal.get(HOUR_OF_DAY))
            fromCalendar.set(MINUTE, cal.get(MINUTE))
            fromCalendar.set(DAY_OF_YEAR, cal.get(DAY_OF_YEAR))

            cal.timeInMillis = roundTime(toCalendar.timeInMillis)
            toCalendar.set(HOUR_OF_DAY, cal.get(HOUR_OF_DAY))
            toCalendar.set(MINUTE, cal.get(MINUTE))
            toCalendar.set(DAY_OF_YEAR, cal.get(DAY_OF_YEAR))
        }

        if (ignoreOverlapping) {
            return Estimate(estimateType, fromCalendar, toCalendar)
        }

        val newRecord = SleepRecord(TimeZone.getDefault().id, fromCalendar.time, toCalendar.time)
        newRecord.to = toCalendar.time
        newRecord.isFinished = true

        val overlappingRecord = findOverlap(newRecord) ?: return Estimate(estimateType, fromCalendar, toCalendar)

        val from = overlappingRecord.from.time
        val to = overlappingRecord.to.time

        Logger.logInfo("Overlap " + overlappingRecord + " FROM " + fromCalendar.time + " TO " + toCalendar.time)

        if (fromCalendar.timeInMillis >= from && toCalendar.timeInMillis <= to) {
            Logger.logInfo("Overlap inside ")
            fromCalendar.timeInMillis = to
            fromCalendar.add(MINUTE, 5)
            toCalendar.timeInMillis = to
            toCalendar.add(HOUR_OF_DAY, 1)
        } else if (fromCalendar.getTimeInMillis() in from..to) {
            Logger.logInfo("Overlap after ")
            fromCalendar.timeInMillis = to
            fromCalendar.add(MINUTE, 5)
            if (toCalendar.timeInMillis - fromCalendar.timeInMillis < 30 * 60000) {
                toCalendar.timeInMillis = fromCalendar.timeInMillis
                toCalendar.add(MINUTE, 30)
            }
        } else if (toCalendar.timeInMillis in from..to) {
            Logger.logInfo("Overlap before ")
            toCalendar.timeInMillis = from
            toCalendar.add(MINUTE, -5)
            if (toCalendar.timeInMillis - fromCalendar.timeInMillis < 30 * 60000) {
                fromCalendar.timeInMillis = toCalendar.timeInMillis
                fromCalendar.add(MINUTE, -30)
            }
        }
        if (from > fromCalendar.timeInMillis && to < toCalendar.timeInMillis) {
            Logger.logInfo("Overlap inside reversed ")
            fromCalendar.timeInMillis = to
            fromCalendar.add(MINUTE, 5)
            toCalendar.timeInMillis = toCalendar.timeInMillis
            if (toCalendar.timeInMillis - fromCalendar.timeInMillis < 30 * 60000) {
                toCalendar.timeInMillis = fromCalendar.timeInMillis
                toCalendar.add(MINUTE, 30)
            }
        }
        return Estimate(estimateType, fromCalendar, toCalendar)
    }

    private fun findOverlap(newRecord: SleepRecord): SleepRecord? {
        val sleepRecords = SleepRecords()
        val recordsAround = SharedApplicationContext.getInstance().sleepRecordRepository
                .getSleepRecords(newRecord.from.time - DAY, newRecord.to.time + DAY, false)
        recordsAround
                ?.filter { sleepRecords.isOverlap(it, newRecord) }
                ?.forEach { return it }
        return null
    }

    private fun roundTime(time: Long): Long {
        val mod = time % 300000
        return time + if (mod > 150000) 300000 - mod else -mod
    }

}