package com.urbandroid.sleep.activityrecognition.calculator

import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.activityrecognition.ActivityIntervals
import com.urbandroid.sleep.activityrecognition.ActivityIntervals.Companion.format
import com.urbandroid.sleep.activityrecognition.calculator.SleepTimeCalculator.Estimate
import java.util.*
import java.util.concurrent.TimeUnit.MILLISECONDS
import java.util.concurrent.TimeUnit.MINUTES


class SleepTimeDetectedActivityCalculator(private val intervals: List<ActivityIntervals.Interval>): SleepTimeCalculator {

    override fun estimate(estimate: Estimate): Estimate {

        val fromCal = estimate.from
        val toCal = estimate.to

        val range = ActivityIntervals.Interval(fromCal.timeInMillis, toCal.timeInMillis)

        // find first and last - filter all with length >= 2 Hours
        val longStillIntervals = intervals
                .filterNot { it.isOpen }
                .filter { (it.from > fromCal.timeInMillis && it.to < toCal.timeInMillis) || (fromCal.timeInMillis in it) || (toCal.timeInMillis in it)}
                .filter {
                    it.length() in MIN_LONG_STILL_INTERVAL_IN_HOURS .. MAX_LONG_STILL_INTERVAL_IN_HOURS
                }
                .filter {
                    // filter those which have intersection with original range at least half of original range length
                    val intersection = it.intersection(range)
                    intersection?.length(MINUTES)?:0 > range.length(MINUTES)/2
                }.filter {
                    // checks
                    // in from time is too before range.from
                    // in to time is too after range.to
                    MILLISECONDS.toHours(range.from - it.from) < 4 && MILLISECONDS.toHours(it.to - range.to) < 4
                }

        Logger.logDebug(
                        "$estimate " +
                        "longStillIntervals: ${ActivityIntervals.from(longStillIntervals).toString(true)}")

        // None still interval found
        if (longStillIntervals.isEmpty()) {
            return Estimate(estimate.type, fromCal, toCal, longStillIntervals)
        }

        // now calculate total length
        val total = longStillIntervals.sumBy { it.length(MINUTES).toInt() }
        if (total < MIN_SLEEP_TIME_IN_HOURS * 60) {
            return Estimate(estimate.type, fromCal, toCal, longStillIntervals)
        }

        val start = Calendar.getInstance().apply {
            timeInMillis = longStillIntervals.first().from
        }

        val end = Calendar.getInstance().apply {
            timeInMillis = with(longStillIntervals) { if (first().length() > 5) first().to else last().to }
        }

        Logger.logDebug("Suggested sleep time: ${format(start.timeInMillis)} to: ${format(end.timeInMillis)}")

        return Estimate(Estimate.Type.ACTIVITY, start, end, longStillIntervals)
    }

    companion object {
        const val MIN_SLEEP_TIME_IN_HOURS = 3
        const val MIN_LONG_STILL_INTERVAL_IN_HOURS = 2
        const val MAX_LONG_STILL_INTERVAL_IN_HOURS = 14
    }
}
