package com.urbandroid.sleep.alarmclock

import android.content.Context
import android.content.Intent
import android.support.v4.app.JobIntentService
import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.persistence.Export
import com.urbandroid.sleep.service.Settings
import com.urbandroid.util.Experiments
import java.io.File
import java.util.concurrent.atomic.AtomicBoolean

class CleanupService : JobIntentService() {

    override fun onHandleWork(intent: Intent) {
        Logger.logInfo("CleanupService.onHandleWork")
        try {
            tasks.forEach {
                if (it.shouldRun()) {
                    it.run()
                }
            }
        }
        catch (e: RuntimeException) {
            Logger.logSevere("CleanupService.onHandleWork", e)
        }
    }

    companion object {

        private const val JOB_ID = 669
        private const val ACTION_CLEANUP = "com.urbandroid.sleep.alarmclock.ACTION_CLEANUP"

        private val startRequested = AtomicBoolean(false)

        private lateinit var tasks: List<CleanupTask>

        @JvmStatic
        fun start(context: Context) {
            try {
                if (startRequested.getAndSet(true)) {
                    return
                }

                Logger.logInfo("CleanupService.start")

                tasks = listOf(
                        DeleteSoundclassLogs(context)
                        //add more tasks as necessary
                )

                val shouldAnyTaskRun = tasks.find { it.shouldRun() } != null

                if (shouldAnyTaskRun) {
                    enqueueWork(context, CleanupService::class.java, JOB_ID, Intent(ACTION_CLEANUP))
                    Logger.logInfo("CleanupService.start: work scheduled")
                } else {
                    Logger.logInfo("CleanupService.start: no tasks")
                }
            }
            catch (e: RuntimeException) {
                Logger.logSevere("CleanupService.start", e)
            }
        }
    }

    interface CleanupTask {

        fun shouldRun(): Boolean

        fun run()
    }

    abstract class OneTimeSettingsBasedTask(val context: Context, val alreadyExecutedSettingsKey: String): CleanupTask {

        val settings = Settings.from(context)

        override fun shouldRun(): Boolean {
            return !settings.getPrefBoolean(alreadyExecutedSettingsKey, false)
        }

        override fun run() {
            try {
                if (shouldRun()) {
                    Logger.logInfo("CleanupService: run task: " + alreadyExecutedSettingsKey)
                    doRun()
                    settings.save(alreadyExecutedSettingsKey, true)
                }
            }
            catch (e: RuntimeException) {
                Logger.logSevere("CleanupService: run task: " + alreadyExecutedSettingsKey, e)
            }
        }

        abstract fun doRun()
    }

    class DeleteSoundclassLogs(context: Context): OneTimeSettingsBasedTask(context, "cleanup_delete_soundclass_logs") {

        override fun shouldRun(): Boolean {
            return !Experiments.getInstance().isDetailedSoundRecognitionLoggingExperiment && super.shouldRun()
        }

        override fun doRun() {
            val extDir = context.getExternalFilesDir(null)
            extDir ?: throw RuntimeException("getExternalFilesDir returned null")
            val logDir = File(extDir!!.absolutePath + Export.EXPORT_PATH)
            if (logDir.isDirectory) {
                val files = logDir.listFiles { dir, name -> name.startsWith("soundclass_") }
                files ?: throw RuntimeException("listFiles returned null")
                files.forEach { it.delete() }
            }
        }
    }
}
