package com.urbandroid.sleep.audio.consumer

import android.content.Context
import android.content.Intent
import android.os.Handler
import android.support.v4.content.LocalBroadcastManager

import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.audio.AudioReadBuffer
import com.urbandroid.sleep.audio.AudioRecorderContext
import com.urbandroid.sleep.audio.event.SoundEvent
import com.urbandroid.sleep.audio.transform.AudioTransformer
import com.urbandroid.sleep.domain.CurrentSleepRecord
import com.urbandroid.sleep.domain.EventLabel
import com.urbandroid.sleep.mic.SnoringReceiver
import com.urbandroid.sleep.snoring.tensorflow.SoundClass
import com.urbandroid.sleep.snoring.tensorflow.SoundHistory
import com.urbandroid.sleep.snoring.tensorflow.TensorflowClassifier
import com.urbandroid.util.ArrayUtil

import com.urbandroid.sleep.sendExplicitBroadcast
import com.urbandroid.sleep.snoring.feature.FftFeature
import com.urbandroid.sleep.snoring.record.*
import com.urbandroid.util.Experiments
import com.urbandroid.util.StringBufferPersister
import java.util.*
import java.util.concurrent.TimeUnit.MILLISECONDS

class EnsembleAudioConsumer(context: Context) : ChunkedAudioConsumer(context, 0) {

    private val decimateFactor = 4

    private val handler = Handler()

    private val chunks = mutableListOf<FloatArray>()

    private val classifier = EnsembleClassifier(context)

    @Volatile
    private var lastAntisnoringTriggereg: Long = 0

    @Volatile
    private var lastSoundEvent: Long = 0

    override fun preprocess(readBuffer: AudioReadBuffer): FloatArray {

        val sampleRate = recorderContext.sampleRate

        val transformer = AudioTransformer.from(sampleRate, readBuffer)
                .iirFilterFactory(recorderContext.iirFilterFactory)
                .lowPass()

        val result = transformer
                .copyData()
                .decimate(decimateFactor)
                .toData()

        return result
    }

    override fun processChunk(data: FloatArray) {

        val now = System.currentTimeMillis()

        chunks.add(data)
        while (chunks.size > classifier.requiredChunksCount) {
            chunks.removeAt(0)
        }
        if (chunks.size < classifier.requiredChunksCount) {
            return
        }

        val soundClass = classifier.classify(chunks)

        LocalBroadcastManager.getInstance(context).sendBroadcast(SoundEvent.of(soundClass).toIntent())

        if (soundClass != SoundClass.OTHER) {

            val maxEventRateSeconds = 30.0
            if (MILLISECONDS.toSeconds(now - lastSoundEvent) >= maxEventRateSeconds) {

                Logger.logDebug("EnsembleAudioConsumer: generating sound event. $soundClass")

                lastSoundEvent = now

                val sleepRecord = CurrentSleepRecord.getInstance().record
                if (sleepRecord != null) {
                    val ts = System.currentTimeMillis() - Math.round(maxEventRateSeconds * 1000)
                    handler.post {
                        if (!sleepRecord.isFinished) {
                            when (soundClass) {
                                SoundClass.SNORE -> {
                                    sleepRecord.addSnore(Math.round(maxEventRateSeconds).toInt())
                                    sleepRecord.addEventLabel(EventLabel.SNORING, ts)
                                }
                                SoundClass.TALK -> sleepRecord.addEventLabel(EventLabel.TALK, ts)
                                SoundClass.BABY -> sleepRecord.addEventLabel(EventLabel.BABY, ts)
                                SoundClass.COUGH -> sleepRecord.addEventLabel(EventLabel.SICK, ts)
                            }
                        }
                    }
                }

                if (soundClass == SoundClass.SNORE && MILLISECONDS.toSeconds(now - lastAntisnoringTriggereg) >= 60) {
                    Logger.logDebug("EnsembleAudioConsumer: antisnoring called")
                    context.sendExplicitBroadcast(Intent(SnoringReceiver.ANTISNORING_ACTION))
                    lastAntisnoringTriggereg = now
                }
            }
        }
    }

    //doOpen/doClose can be called repeatedly.
    //When tracking is paused/resumed explicitly by the user, doClose/doOpen is called, rather than doPause/doResume.

    override fun doOpen(audioRecorderContext: AudioRecorderContext) {
        super.doOpen(audioRecorderContext)
        chunkSize = audioRecorderContext.sampleRate / decimateFactor
        classifier.open()
    }

    override fun doClose() {
        super.doClose()
        chunks.clear()
        classifier.close()
    }

    override fun doPause() {
        super.doPause()
        chunks.clear()
        classifier.pause()
    }

    override fun doResume() {
        super.doResume()
        classifier.resume()
    }

    interface Classifier {

        val requiredChunksCount: Int

        //The chunks are supposed to be 1s long.
        fun classify(chunks: List<FloatArray>): SoundClass

        fun open()

        fun close()

        fun pause()

        fun resume()
    }

    class EnsembleClassifier(private val context: Context) : Classifier {

        @Volatile
        private var tfClassifier: TensorflowClassifier? = TensorflowClassifier(context)

        private val tfHistory = SoundHistory()

        private val slowPeriod = SlowPeriodClassifier()

        private val fileLog : StringBufferPersister? =
                if (Experiments.getInstance().isDetailedSoundRecognitionLoggingExperiment()) {
                    StringBufferPersister("soundclass", 1000, true)
                } else {
                    null
                }

        override val requiredChunksCount = 30

        override fun classify(chunks: List<FloatArray>): SoundClass {

            val sampleRate = chunks[0].size

            val tfSample = ArrayUtil.join(chunks.subList(chunks.size - 3, chunks.size))
            val tfResult = tfClassifier!!.classify(tfSample, sampleRate)

            tfHistory.add(tfResult)
            val tfHistoryResult = tfHistory.classifyRecentHistory()

            val spSample = ArrayUtil.join(chunks)
            val spResult = slowPeriod.classify(spSample, sampleRate)
            val spResultClass = if (spResult >= 0.025) SoundClass.SNORE else SoundClass.OTHER;

            val ensembleResult =
                    if (tfHistoryResult == SoundClass.TALK) {
                        SoundClass.TALK
                    }
                    else if (SoundHistory.containsSnore(tfHistory.history, 2) && spResultClass == SoundClass.SNORE) {
                        SoundClass.SNORE
                    }
                    else {
                        SoundClass.OTHER
                    }

            fileLog?.update("${Date()},$tfResult,$tfHistoryResult,$spResult,$spResultClass,$ensembleResult\n")

            return ensembleResult
        }

        override fun open() {
            if (tfClassifier == null) {
                tfClassifier = TensorflowClassifier(context)
            }
        }

        override fun close() {
            tfHistory.clear()
            if (tfClassifier != null) {
                tfClassifier!!.close()
                tfClassifier = null
            }
            fileLog?.flush()
        }

        override fun pause() {
            tfHistory.clear()
            fileLog?.flush()
        }

        override fun resume() {}
    }

    class SlowPeriodClassifier {

        fun classify(sample: FloatArray, sampleRate: Int): Double {

            var record = Record()
            record.set("SAMPLE", MonoSample(sample, sampleRate))
            record = DownsampleAndMovingMax("SAMPLE_LF", 32, 0.2).apply(record)
            record = ComputeFft("SAMPLE_LF", "FFT_LF").apply(record)

            record = FftFeature.energyMaxRel("FFT_LF", 0.2, 0.5).apply(record)
            val breathFreqEnergy = record.getLastFeature<Double>()

            return breathFreqEnergy
        }
    }
}
