package com.urbandroid.sleep.fragment.dashboard

import android.app.Activity
import com.getkeepsafe.taptargetview.TapTarget
import com.getkeepsafe.taptargetview.TapTargetView
import com.urbandroid.common.releasenotes.ReleaseNotes
import com.urbandroid.common.util.DateUtil
import com.urbandroid.sleep.R
import com.urbandroid.sleep.TrialFilter
import com.urbandroid.sleep.alarmclock.Alarms
import com.urbandroid.sleep.alarmclock.SleepStarter
import com.urbandroid.sleep.domain.Noise
import com.urbandroid.sleep.domain.SleepRecord
import com.urbandroid.sleep.domain.addon.AddonApp
import com.urbandroid.sleep.domain.tag.Tag
import com.urbandroid.sleep.fragment.dashboard.card.*
import com.urbandroid.sleep.service.Settings
import com.urbandroid.sleep.service.SharedApplicationContext
import java.util.concurrent.TimeUnit

class DashboardCardBuilder {

    var lastSleepRecord : SleepRecord? = null

    val noises = mutableListOf<Noise>()

    var interestingNoises = mutableListOf<Noise>()

    var lastNoise : Noise? = null

    var cachedCards = mutableListOf<DashboardCard<*>>()

    // in background
    fun loadData() {
        val db = SharedApplicationContext.getInstance().sleepRecordRepository
        lastSleepRecord = db.lastSleepRecord

        with(db.getNoises(1)) {
            if (this.size > 0) {
                lastNoise = this.get(0)
            }
        }

        lastSleepRecord?.let {

            noises.clear()
            noises.addAll(db.getNoises(100, it.fromTime, it.toTime))

            val current = it

            interestingNoises.apply {
                clear()
                addAll(noises.filter {
                    TimeUnit.MILLISECONDS.toMinutes(it.fromTime - current.fromTime) > 15 &&
                            TimeUnit.MILLISECONDS.toMinutes(current.toTime - it.toTime) > 15 &&
                            (Tag.hasTag(it.comment, Tag.TALK) || Tag.hasTag(it.comment, Tag.SNORE))
                })
            }

            interestingNoises.sortWith(compareBy({!Tag.hasTag(it.comment, Tag.TALK)}, {it.fromTime}))

//            for (noise in interestingNoises) {
//                Logger.logInfo("DCB noise " + noise.comment)
//            }
        }
    }

    // on main thread
    fun buildCards(context: Activity): List<DashboardCard<*>> {
        cachedCards.clear()

        var cards = mutableListOf<DashboardCard<*>>()

        val settings = Settings(context)

        if (settings.isShowReleaseNotes) {
            if (ReleaseNotes(context).isNewVersion) {
                cards.add(ReleaseNotesCard(context))
            }
        }

        val bedtimeCard = BedtimeCard(context)

        cards.add(bedtimeCard)
        bedtimeCard.setOnCardClickedListener {
            if (settings.nextAlarm != null) {
                val minutesOfSleep = Math.round(((settings.nextAlarm.time - System.currentTimeMillis()) / 60000).toFloat())
                val text = context.resources.getString(R.string.sleep) + " " + DateUtil.formatMinutesInHumanLanguage(context, minutesOfSleep)
                TapTargetView.showFor(context,
                        TapTarget.forView(context.findViewById(R.id.fab), context.resources.getString(R.string.start_sleep_tracking), text)
                                .dimColor(R.color.transparent_black)
                                .outerCircleColor(R.color.t4)
                                .targetCircleColor(R.color.white)
                                .cancelable(true)
                                .tintTarget(false)
                                .textColor(R.color.white), object : TapTargetView.Listener() {

                    override fun onTargetClick(view: TapTargetView) {
                        super.onTargetClick(view)
                        SleepStarter().startSleep(context)
                    }
                })
            }
        }

        var isRecordRecent = false
        lastSleepRecord?.let {
            isRecordRecent = System.currentTimeMillis() - it.toTime < 60000 * 60 * 4
        }

        val nextAlarm = Alarms.calculateNextAlert(context)
        cards.add(NextAlarmCard(context, nextAlarm))

        if (TrialFilter.getInstance().isTrial) {
//            if (TrialFilter.getInstance().daysUsed() > 0) {
                cards.add(UnlockCard(context))
                if (((TrialFilter.getInstance().isTrialExpired) || (TrialFilter.getInstance().version == TrialFilter.Version.FULLAD))) {
                    cards.add(NativeAdCard(context))
                }
//            }
        }

        if (interestingNoises.isNotEmpty()) {
            cards.add(InterestingNoiseCard(context, interestingNoises))
        }

        if (lastSleepRecord != null && isRecordRecent) {
            cards.add(LastSleepRecordCard(context, lastSleepRecord))
        }

        val sleepScore = Settings(context).sleepScore
        cards.add(ScoreCard(context, sleepScore))
        cards.add(GoalCard(context))
        
        if (lastSleepRecord != null) {
            if (!isRecordRecent) {
                cards.add(LastSleepRecordCard(context, lastSleepRecord))
            }

            if (lastNoise == null) {
                if (!settings.isRecordingMasterSwitchEnabled) {
                    cards.add(NoNoiseCard(context))
                }
            } else if (noises.size > 0) {
                cards.add(NoiseRecordCard(context, lastSleepRecord, noises))
            }
        } else {
            cards.add(BackupCard(context))
            cards.add(NoSleepRecordCard(context))
        }

        if (settings.recordsCount >= 7) {
            cards.add(ChartCard(context))
        }

        cards.add(AdviceCard(context))

        cards.add(ShortcutCard(context))


        val addonApp = AddonApp.random(context)
        if (lastSleepRecord != null && addonApp != null) {
            cards.add(AddOnCard(context, addonApp))
        }

        val pinnedCards = cards.filter { it.isPinned }
        val notPinnedCards = cards.filter { !it.isPinned }

        val lastCard = LastCard(context, cards.toList())

        cachedCards = cachedCards.apply {
                    addAll(pinnedCards)
                    addAll(notPinnedCards)
                    add(lastCard)
                }
                .filter { !it.isRemoved }.toMutableList()

        cachedCards.map {
            it.setCardBuilder(this)
        }


        return cachedCards.toList()
    }

    fun findPosition(context: Activity, card: DashboardCard<*>): Int? {
        if (cachedCards.size == 0) {
            buildCards(context)
        }
        return cachedCards.indices.firstOrNull { cachedCards[it].type == card.type }
    }

}