package com.urbandroid.sleep.nearby.core

import android.content.Context
import android.os.Handler
import com.google.android.gms.nearby.Nearby
import com.google.android.gms.nearby.connection.*
import java.io.Closeable
import java.util.*

/**
 * A simple wrapper around nearby connections API.
 *
 * * It automatically connects to one endpoint on the same service name,
 *   and it periodically attempts to reconnect if anything goes wrong.
 *
 * * Everything runs explicitly in the main thread. I used to experience weird errors
 *   when the nearby API calls were invoked from a different thread.
 *
 * * send() does not guarantee anything, and it does not throw any exception if things go wrong.
 *   It is just a reflection of the underlying API behaviour.
 *   It also sometimes just fails silently and does not deliver the playload.
 */
class Endpoint(private val serviceName: String,
               private val context: Context,
               private val logger: MyLogger,
               private val payloadConsumer: (ByteArray) -> Unit)
    : Closeable {

    private val mainHandler = Handler(context.mainLooper)

    private val strategy = Strategy.P2P_POINT_TO_POINT

    private val recoveryIntervalMillis = 5000L

    private val endpointName = UUID.randomUUID().toString()

    private val client : ConnectionsClient

    @Volatile
    private var needsRecovery = false

    @Volatile
    private var lastRecoveryAttempt = System.currentTimeMillis()

    @Volatile
    private var closed = false

    @Volatile
    private var remoteEndpointId: String? = null

    init {
        checkCoarseLocationPermission(context)
        client = Nearby.getConnectionsClient(context)
        startAdvertising()
        startDiscovery()
        mainHandler.postDelayed(RecoverStep1(), recoveryIntervalMillis)
    }

    private fun startAdvertising() {

        logger.logInfo("Nearby: startAdvertising")

        client.startAdvertising(
                endpointName, serviceName, MyConnectionLifecycleCallback(),
                AdvertisingOptions.Builder().setStrategy(strategy).build()
        )
        .addOnSuccessListener {
            logger.logInfo("Nearby: startAdvertising success")
        }
        .addOnFailureListener {
            logger.logInfo("Nearby: startAdvertising failure", it)
            needsRecovery = true
        }
    }

    private fun startDiscovery() {

        logger.logInfo("Nearby: startDiscovery")

        client.startDiscovery(
                serviceName,
                object: EndpointDiscoveryCallback() {

                    override fun onEndpointFound(endpointId: String, epInfo: DiscoveredEndpointInfo) {

                        logger.logInfo("Nearby: onEndpointFound $endpointId ${epInfo.serviceId} ${epInfo.endpointName}")

                        if (serviceName == epInfo.serviceId) {
                            if (!isConnected()) {
                                if (epInfo.endpointName < endpointName) {
                                    requestConnection(endpointId)
                                } else {
                                    requestConnectionAfterDelay(endpointId)
                                }
                            } else {
                                logger.logWarning("Nearby: already connected, ignoring the new device.")
                            }
                        }
                    }

                    override fun onEndpointLost(endpointId: String) {
                        logger.logInfo("Nearby: onEndpointLost $endpointId")
                    }
                },
                DiscoveryOptions.Builder().setStrategy(strategy).build()
        )
        .addOnSuccessListener {
            logger.logInfo("Nearby: startDiscovery success")
        }
        .addOnFailureListener {
            logger.logInfo("Nearby: startDiscovery failure", it)
            needsRecovery = true
        }
    }

    private fun requestConnection(discoveredEndpointId: String) {
        logger.logInfo("Nearby: requestConnection $discoveredEndpointId")
        val task = client.requestConnection(endpointName, discoveredEndpointId, MyConnectionLifecycleCallback())
        task.addOnSuccessListener {
            logger.logInfo("Nearby: requestConnection success")
        }
        task.addOnFailureListener {
            logger.logInfo("Nearby: requestConnection failure", it)
            needsRecovery = true
        }
    }

    private fun requestConnectionAfterDelay(discoveredEndpointId: String) {
        logger.logInfo("Nearby: ready to accept connection")
        // Normally, the both nodes see each other, so the one with "greater" name initiates
        // the connection to the one with the "lower" name.
        // But, somehow, sometimes, it happens that only one node sees the other,
        // and if it is the one with lower ID, it would wait for the incoming connection forever.
        // So, if the connection is not established in 30 seconds,
        // the node with "lower" name attempts to initiate it.
        mainHandler.postDelayed(
                {if (!isConnected()) {
                    logger.logInfo("Nearby: No incoming connection yet, trying to initiate it.")
                    requestConnection(discoveredEndpointId)
                }},
                30000)
    }

    fun send(payload: ByteArray) {
        mainHandler.post({
            if (!(closed || needsRecovery || remoteEndpointId == null)) {
                client.sendPayload(remoteEndpointId!!, Payload.fromBytes(payload))
                .addOnSuccessListener {
                    logger.logInfo("Nearby: sendPayload success")
                }
                .addOnFailureListener {
                    logger.logInfo("Nearby: sendPayload failure", it)
                }
            }
        })
    }

    private fun shutdown() {
        logger.logInfo("Nearby: shutdown")
        remoteEndpointId = null
        client.stopDiscovery()
        client.stopAdvertising()
        client.stopAllEndpoints()
    }

    override fun close() {
        mainHandler.post({
            closed = true
            shutdown()
            logger.logInfo("Nearby: close")
        })
    }

    fun isConnected(): Boolean {
        return remoteEndpointId != null
    }

    fun isClosed(): Boolean {
        return closed
    }

    inner class MyConnectionLifecycleCallback(): ConnectionLifecycleCallback() {

        override fun onConnectionInitiated(endpointId: String, connInfo: ConnectionInfo) {
            logger.logInfo("Nearby: onConnectionInitiated $endpointId ${connInfo.endpointName}")
            client.acceptConnection(endpointId, MyPayloadCallback())
        }

        override fun onConnectionResult(endpointId: String, resolution: ConnectionResolution)  {
            logger.logInfo("Nearby: onConnectionResult $endpointId ${resolution.status}")
            if (resolution.status.isSuccess()) {
                remoteEndpointId = endpointId
                client.stopDiscovery()
                client.stopAdvertising()
            } else {
                needsRecovery = true
            }
        }

        override fun onDisconnected(endpointId: String) {
            logger.logInfo("Nearby: onDisconnected $endpointId")
            needsRecovery = true
        }
    }

    inner class MyPayloadCallback : PayloadCallback() {

        private var payload: Payload? = null

        override fun onPayloadReceived(endpointId: String, payload: Payload) {
            logger.logInfo("Nearby: onPayloadReceived ${payload.id}")
            this.payload = payload
        }

        override fun onPayloadTransferUpdate(endpointId: String, update: PayloadTransferUpdate) {
            logger.logInfo("Nearby: onPayloadTransferUpdate: ${update.payloadId} ${update.status} ${update.bytesTransferred}")
            if (payload == null || payload!!.id != update.payloadId) {
                //sender
                when (update.status) {
                    PayloadTransferUpdate.Status.SUCCESS -> {}
                    PayloadTransferUpdate.Status.CANCELED, PayloadTransferUpdate.Status.FAILURE -> {
                        needsRecovery = true
                    }
                }
            }
            else {
                //receiver
                when (update.status) {
                    PayloadTransferUpdate.Status.SUCCESS -> {
                        payloadConsumer(payload!!.asBytes()!!)
                        payload = null;
                    }
                    PayloadTransferUpdate.Status.CANCELED, PayloadTransferUpdate.Status.FAILURE -> {
                        needsRecovery = true
                        payload = null
                    }
                }
            }
        }
    }

    inner class RecoverStep1: Runnable {
        override fun run() {
            if (!closed) {

                val sinceLastRecovery = System.currentTimeMillis() - lastRecoveryAttempt
                if (sinceLastRecovery > 2*60*1000 && !isConnected())  {
                    // Advertising and discovery started more than 2 minutes ago,
                    // and no connection has been established yet.
                    // Sometimes the advertising/discovery process gets stuck, and restart may help.
                    needsRecovery = true
                }

                if (needsRecovery) {
                    // Just shutdown and schedule the actual reconnect for later.
                    // When reconnect follows immediately, it fails more often.
                    // The internals of nearby API probably need some time to cleanup asynchronously.
                    shutdown()
                    mainHandler.postDelayed(RecoverStep2(), recoveryIntervalMillis)
                }
                else {
                    mainHandler.postDelayed(RecoverStep1(), recoveryIntervalMillis)
                }
            }
        }
    }

    inner class RecoverStep2: Runnable {
        override fun run() {
            if (!closed) {

                needsRecovery = false;
                lastRecoveryAttempt = System.currentTimeMillis()

                startAdvertising()
                startDiscovery()

                mainHandler.postDelayed(RecoverStep1(), recoveryIntervalMillis)
            }
        }
    }
}
