package com.urbandroid.sleep.nearby.core

import android.content.Context
import android.os.Handler
import org.json.JSONArray
import org.json.JSONObject
import java.io.Closeable
import kotlin.math.min

/**
 * Reliable wrapper around Endpoint.
 * It buffers outgoing messages and keeps resending them until the counterparty acknowledges them.
 * It periodically acknowledges received messages and removes duplicates.
 * For simplicity, everything runs in the main thread. There is some json parsing/serialization,
 * let's see if it causes any performance problems. It can be moved to a background thread.
 */
class ReliableMessageQueue(private val serviceName: String,
                           private val context: Context,
                           private val logger: MyLogger,
                           private val messageConsumer: (Message) -> Unit,
                           private val maxBufferSize: Int = 1000,
                           private val flushIntervalMillis: Long = 10000)
    : Closeable {

    private val mainHandler = Handler(context.mainLooper)

    private val endpoint = Endpoint(serviceName, context, logger, { consume(it) })

    private val buffer = mutableListOf<Message>()

    private var serialNo = 0L

    private var lastReceivedSerialNo = -1L

    init {
        mainHandler.post(PeriodicSendAndAcknowledge())
    }

    private fun consume(byteArray: ByteArray) {
        val jsonString = String(byteArray)
        logger.logDebug("Nearby: received json: $jsonString")
        val jsonRoot = JSONObject(jsonString)
        val type = jsonRoot.getString("type")
        when (type) {
            "msg" -> consumeMessages(jsonRoot.getJSONArray("val"))
            "unr" -> consumeUnreliableMessage(jsonRoot.getJSONObject("val"))
            "ack" -> consumeAck(jsonRoot.getLong("val"))
            else -> logger.logWarning("Unknown type: $type")
        }
    }

    private fun consumeMessages(jsonMessages: JSONArray) {
        for(i in 0 until jsonMessages.length()) {
            val jsonMessage = jsonMessages.getJSONObject(i)
            val serialNumber = jsonMessage.getLong("sn")
            val payload = jsonMessage.getJSONObject("pl")
            if (serialNumber > lastReceivedSerialNo) {
                messageConsumer(Message(serialNumber, payload))
                lastReceivedSerialNo = serialNumber
            }
        }
    }

    private fun consumeUnreliableMessage(payload: JSONObject) {
        messageConsumer(Message(-1, payload))
    }

    private fun consumeAck(ack: Long) {
        while(!buffer.isEmpty() && buffer.get(0).serialNo <= ack) {
            buffer.removeAt(0)
        }
    }

    /**
     * Add the data to the buffer, keep sending them until they get acknowledged.
     */
    fun sendReliable(payload: JSONObject) {
        mainHandler.post({
            // TODO I should restrict the buffer size by total memory consumption of the buffered
            // messages, rather that by message count. But it is obviously complicated.
            // At the moment, it relies on reasonable usage by the client code:
            // it needs to send reasonably small messages for the given buffer size.
            while(buffer.size >= maxBufferSize) {
                buffer.removeAt(0)
            }
            buffer.add(Message(serialNo, payload))
            serialNo += 1
        })
    }

    /**
     * Send the data immediately, no acknowledgement, no guarantees.
     */
    fun sendUnreliable(payload: JSONObject) {
        val jsonRoot = JSONObject();
        jsonRoot.put("type", "unr")
        jsonRoot.put("val", payload)
        val jsonString = jsonRoot.toString()
        logger.logDebug("Nearby: sending json: $jsonString")
        val byteArray = jsonString.toByteArray()
        endpoint.send(byteArray)
    }

    override fun close() {
        endpoint.close()
    }

    inner class PeriodicSendAndAcknowledge : Runnable {
        override fun run() {
            if (!endpoint.isClosed()) {
                flush()
                mainHandler.postDelayed(this, flushIntervalMillis)
            }
        }
    }

    private fun flush() {
        if (endpoint.isClosed() || !endpoint.isConnected() || buffer.isEmpty()) {
            return
        }
        sendAck()
        sendMessages()
    }

    private fun sendAck() {
        val jsonRoot = JSONObject();
        jsonRoot.put("type", "ack")
        jsonRoot.put("val", lastReceivedSerialNo)
        val jsonString = jsonRoot.toString()
        logger.logDebug("Nearby: sending json: $jsonString")
        val byteArray = jsonString.toByteArray()
        endpoint.send(byteArray)
    }

    private fun sendMessages() {
        // TODO Max payload size is 32 kb.
        // I should check the actual json size, rather than the number of messages.
        // How about zipping the payload?
        val maxMessagesPerBatch = 100
        val jsonMessages = JSONArray()
        for (i in 0 until min(maxMessagesPerBatch, buffer.size)) {
            val message = buffer.get(i)
            val jsonMessage = JSONObject()
            jsonMessage.put("sn", message.serialNo)
            jsonMessage.put("pl", message.payload)
            jsonMessages.put(jsonMessage)
        }
        val jsonRoot = JSONObject();
        jsonRoot.put("type", "msg")
        jsonRoot.put("val", jsonMessages)
        val jsonString = jsonRoot.toString()
        logger.logDebug("Nearby: sending json: $jsonString")
        val byteArray = jsonString.toByteArray()
        endpoint.send(byteArray)
    }
}
