package com.urbandroid.sleep.nearby.pairtracking

import android.content.Context
import android.os.Handler
import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.nearby.core.Message
import com.urbandroid.sleep.nearby.core.MyLogger
import com.urbandroid.sleep.nearby.core.ReliableMessageQueue
import com.urbandroid.sleep.nearby.core.getServiceName
import com.urbandroid.sleep.snoring.feature.Moving
import com.urbandroid.util.StringBufferPersister
import org.json.JSONObject
import java.io.Closeable
import java.util.*

class PairActivityTracking(private val ctx: Context): Closeable {

    private val mainHandler = Handler(ctx.mainLooper)

    private val mq: ReliableMessageQueue

    private val dataLog = StringBufferPersister("nearby_data", 100, true)
    private val detailsLog = StringBufferPersister("nearby_details", 1000, true)

    private val minTimeDiff = Moving.min(15)

    private val rawPairActivity = RawPairActivity()

    init {
        //TODO use a shared secret token from Settings
        val serviceName = getServiceName("PairActivityTracking")

        val logger = MyLogger { level, message, error ->
            if (level > Logger.DEBUG_LEVEL) {
                val errorMessage = if (error == null) "" else "${error.javaClass.simpleName} ${error.message}"
                detailsLog.update("${Date()} $message $errorMessage\n")
            }
        }

        mq = ReliableMessageQueue(serviceName, ctx, logger, { consume(it) })

        val sendCurrentTimeRate = 60000L
        mainHandler.postDelayed(
                object: Runnable {
                    override fun run() {
                        val json = JSONObject()
                        json.put("t", getLocalTime())
                        mq.sendUnreliable(json);
                        mainHandler.postDelayed(this, sendCurrentTimeRate)
                    }
                },
                sendCurrentTimeRate)

        instance = this
    }

    private fun consume(m: Message) {
        if (m.serialNo < 0) {
            val remoteTime = m.payload.getLong("t")
            val localTime = getLocalTime()
            val timeDiff = minTimeDiff.apply((localTime - remoteTime).toFloat()).toLong()
            dataLog.update("time,$localTime,$remoteTime,$timeDiff\n")
            rawPairActivity.addTimeDiff(localTime, timeDiff)
        }
        else {
            val serialNo = m.serialNo
            val eventTime = m.payload.getLong("t")
            val activity = m.payload.getDouble("a")
            //rmtActigraph.append(eventTime, activity.toFloat())
            dataLog.update("remt,$eventTime,$activity,$serialNo\n")
            rawPairActivity.addRemoteActivity(eventTime, activity.toFloat())
        }
    }

    fun update(activity: Double) {
        val json = JSONObject()
        val localTime = getLocalTime()
        json.put("t", localTime)
        json.put("a", activity)
        //locActigraph.append(localTime, activity.toFloat())
        mq.sendReliable(json)
        dataLog.update("mine,$localTime,$activity\n")
        rawPairActivity.addLocalActivity(localTime, activity.toFloat())
    }

    private fun getLocalTime() = System.currentTimeMillis()

    fun getCleanAggregatedActigraph(): FloatArray? {

        //Don't do the pair normalization if the sleep is shorter than 1 hour.
        //TODO some more criteria, e.g. there are enough data from the remote phone.
        if (rawPairActivity.localHistorySize() < 60*6) {
            return null
        }

        val pairActivity = rawPairActivity.normalizeAmplitudes().matchTimes()
        val aggrPairActivity = pairActivity.aggregate(30)
        val result = aggrPairActivity.subtractRemoteActivity(false)
        return result
    }

    override fun close() {
        mq.close()
        dataLog.flush()
        detailsLog.flush()
        instance = null
    }

    //TODO ugly hack, just for now, to make the data visible for the rest of the app.
    companion object {

        @Volatile private var instance: PairActivityTracking? = null

        fun getCleanAggregatedActigraph(): FloatArray? {
            return instance?.getCleanAggregatedActigraph() ?: null
        }
    }
}
