package com.urbandroid.sleep.nearby.pairtracking

import com.urbandroid.sleep.domain.AdaptiveNormalizationFilter
import kotlin.math.abs

class RawPairActivity(
        private val local: LFSeries = LFSeries(),
        private val remote: LFSeries = LFSeries(),
        private val timeDiff: LLSeries = LLSeries()) {

    //TODO we also need to normalize the time points (smartwatch in batch mode feeds like
    //10 values at a time, we need to re-introduce the 10 sec time gaps).
    //This can be done by normalizeTimeSpacing (see below).
    //SleepPhaser is even more tricky - we fetch the valuers every 17 sec, and we need
    //to estimate when the actual intervals start within the sleepphaser.

    fun normalizeAmplitudes(): RawPairActivity {
        return RawPairActivity(
                LFSeries(local.keysAsLongArray(),
                        AdaptiveNormalizationFilter.normalizeAmplitudes(local.valuesAsFloatArray()).output),
                LFSeries(remote.keysAsLongArray(),
                        AdaptiveNormalizationFilter.normalizeAmplitudes(remote.valuesAsFloatArray()).output),
                timeDiff
        )
    }

    fun matchTimes(): PairActivity {

        val lTimes = local.copyOfKeys()
        val lValues = local.copyOfValues()
        val rTimes = remote.copyOfKeys()
        val rValues = remote.copyOfValues()
        val adjRValues = FloatAppendableArray(initSize = lTimes.size())

        for(i in 0 until lTimes.size()) {

            val lTime = lTimes.get(i)
            val rTimeEstimate = lTime - timeDiff.getLGE(lTime) //TODO is it really minus?

            //TODO Now I simply take the closest remote value in time.
            //Interpolation of two or more closest values might be better.
            val rtiLE = rTimes.findLE(rTimeEstimate)
            val rtiGE = rTimes.findGE(rTimeEstimate)
            val rtDiffLE = if (rtiLE < 0) Long.MAX_VALUE else abs(lTime - rTimes.get(rtiLE))
            val rtDiffGE = if (rtiGE >= rTimes.size()) Long.MAX_VALUE else abs(lTime - rTimes.get(rtiGE))
            val maxDistance = 10000L
            val rti =
                    if (rtDiffLE < maxDistance) {
                        if (rtDiffLE < rtDiffGE) rtiLE else rtiGE
                    }
                    else if (rtDiffGE < maxDistance) {
                        if (rtDiffLE < rtDiffGE) rtiLE else rtiGE
                    }
                    else {
                        -1
                    }
            val rv = if (rti < 0) 0f else rValues.get(rti)
            adjRValues.add(rv)
        }

        return PairActivity(lTimes, lValues, adjRValues)
    }

    fun addLocalActivity(time: Long, activity: Float) {
        local.add(adjustNewKey(local, time), activity)
    }

    fun addRemoteActivity(time: Long, activity: Float) {
        remote.add(adjustNewKey(remote, time), activity)
    }

    fun addTimeDiff(time: Long, diff: Long) {
        timeDiff.add(adjustNewKey(timeDiff, time), diff)
    }

    private fun adjustNewKey(series: Series<Long,out Any>, newKey: Long): Long {
        val lastKeyOrNull = series.lastKeyOrNull()
        if (lastKeyOrNull == null || lastKeyOrNull < newKey) {
            return newKey
        } else {
            return lastKeyOrNull + 1
        }
    }

    fun getLocalActivity(): FloatArray {
        return local.valuesAsFloatArray()
    }

    fun getRemoteActivity(): FloatArray {
        return remote.valuesAsFloatArray()
    }

    fun localHistorySize(): Int {
        return local.size()
    }
}

/*
//When data come in a batch, they have the same (or nearly the same) timestamp.
//This method renews regular 10 spacing between the data points.
fun normalizeTimeSpacing(src: LongArray, step: Long = 10000): LongArray {
    val result = src.copyOf()
    val threshold = step / 2
    for(i in result.size-2 downTo 0) {
        if (result[i+1] - result[i] < threshold) {
            result[i] = result[i+1] - step
        }
    }
    return result
}

@Test
fun testNormalizeTimeSpacing() {
    assertThat(normalizeTimeSpacing(longArrayOf(10000, 19000, 39000))).containsExactly(10000, 19000, 39000)
    assertThat(normalizeTimeSpacing(longArrayOf(1,2,3))).containsExactly(-19997, -9997, 3)
    assertThat(normalizeTimeSpacing(longArrayOf(20000, 20001, 40003, 40003))).containsExactly(10001, 20001, 30003, 40003)
}
*/
