package com.urbandroid.sleep.nearby.pairtracking

interface MyArray<T> {

    fun size(): Int

    fun isEmpty(): Boolean

    fun get(i: Int): T

    fun getLastOrNull(): T?

    fun toArray(): Array<T>

    fun toPrimitiveArray(): Any

    fun copyOf(): MyArray<T>
}

interface AppendableArray<T>: MyArray<T> {

    fun add(x: T)

    override fun copyOf(): AppendableArray<T>
}

interface AscSortedArray<T>: MyArray<T> {

    fun findLE(x: T): Int

    fun findGE(x: T): Int

    override fun copyOf(): AscSortedArray<T>
}

interface AscSortedAppendableArray<T>: AppendableArray<T>, AscSortedArray<T> {

    override fun copyOf(): AscSortedAppendableArray<T>
}

abstract class ArrayBase<T>: MyArray<T> {

    override fun isEmpty() = size() == 0

    override fun getLastOrNull() = if (isEmpty()) null else get(size() - 1)
}

open class LongAppendableArray(initValues: LongArray = longArrayOf(), initSize: Int = 1024): ArrayBase<Long>(), AppendableArray<Long> {

    var size: Int
        private set

    protected var values: LongArray
        private set

    init{
        if (initValues.isEmpty()) {
            if (initSize < 1) {
                throw IllegalArgumentException(initSize.toString())
            }
            values = LongArray(initSize)
            size = 0
        } else {
            values = initValues
            size = initValues.size
        }
    }

    override fun size() = size

    override fun get(i: Int): Long {
        if (i < 0 || i >= size) {
            throw IndexOutOfBoundsException("i: $i, size: $size")
        }
        return values[i]
    }

    override fun add(x: Long) {
        if (size == values.size) {
            val oldValues =  values
            values = LongArray(((size+1)*1.5).toInt())
            System.arraycopy(oldValues, 0, values, 0, oldValues.size)
        }
        values[size] = x
        size += 1;
    }

    override fun toString() = joinToString()

    fun joinToString(separator: String = ", ", prefix: String = "[", postfix: String = "]")
            = toPrimitiveArray().joinToString(separator, prefix, postfix)

    override fun toArray() = toPrimitiveArray().toTypedArray()

    override fun toPrimitiveArray() = values.sliceArray(0 until size)

    override fun copyOf() = LongAppendableArray(initValues = toPrimitiveArray())
}

class FloatAppendableArray(initValues: FloatArray = floatArrayOf(), initSize: Int = 1024): ArrayBase<Float>(), AppendableArray<Float> {

    var size: Int
        private set

    protected var values: FloatArray
        private set

    init{
        if (initValues.isEmpty()) {
            if (initSize < 1) {
                throw IllegalArgumentException(initSize.toString())
            }
            values = FloatArray(initSize)
            size = 0
        } else {
            values = initValues
            size = initValues.size
        }
    }

    override fun size() = size

    override fun get(i: Int): Float {
        if (i < 0 || i >= size) {
            throw IndexOutOfBoundsException("i: $i, size: $size")
        }
        return values[i]
    }

    override fun add(x: Float) {
        if (size == values.size) {
            val oldValues =  values
            values = FloatArray(((size+1)*1.5).toInt())
            System.arraycopy(oldValues, 0, values, 0, oldValues.size)
        }
        values[size] = x
        size += 1;
    }

    override fun toString() = joinToString()

    fun joinToString(separator: String = ", ", prefix: String = "[", postfix: String = "]")
            = toPrimitiveArray().joinToString(separator, prefix, postfix)

    override fun toArray() = toPrimitiveArray().toTypedArray()

    override fun toPrimitiveArray() = values.sliceArray(0 until size)

    override fun copyOf() = FloatAppendableArray(initValues = toPrimitiveArray())
}

class LongAscSortedAppendableArray(initValues: LongArray = longArrayOf(), initSize: Int = 1024):
        LongAppendableArray(initValues, initSize), AscSortedAppendableArray<Long> {

    init {
        for(i in 1 until size()) {
            if (get(i-1) >= get(i)) {
                throw IllegalArgumentException("Not in ascending order: $i ${get(i-1)} ${get(i)}")
            }
        }
    }

    override fun add(x: Long) {
        if (size > 0 && x <= get(size-1)) {
            throw IllegalArgumentException("Ordering violation: $x < ${get(size-1)}")
        }
        super.add(x)
    }

    override fun findLE(x: Long): Int {
        val i = values.binarySearch(x, 0, size)
        if (i < 0) {
            return -i - 2
        } else {
            return i
        }
    }

    override fun findGE(x: Long): Int {
        val i = values.binarySearch(x, 0, size)
        if (i < 0) {
            return -i - 1
        } else {
            return i
        }
    }

    override fun copyOf() = LongAscSortedAppendableArray(initValues = toPrimitiveArray())
}

open class Series<K,V>(
        protected val defaultValue: V,
        protected val keys: AscSortedAppendableArray<K>,
        protected val values: AppendableArray<V>) {

    init {
        if (keys.size() != values.size()) {
            throw IllegalArgumentException("${keys.size()} != ${values.size()}")
        }
    }

    fun size() = keys.size()

    fun isEmpty() = size() == 0

    fun add(k: K, v: V) {
        keys.add(k)
        values.add(v)
    }

    fun key(i: Int) = keys.get(i)

    fun value(i: Int) = values.get(i)

    fun lastKeyOrNull() = keys.getLastOrNull()

    fun lastValueOrNull() = values.getLastOrNull()

    /**
     * Find the closest key lower or equal k and return the corresponding value
     */
    fun getLE(k: K) : V {
        val i = keys.findLE(k)
        if (i >= 0) {
            return value(i)
        } else {
            return defaultValue;
        }
    }

    /**
     * Find the closest key greater or equal k and return the corresponding value
     */
    fun getGE(k: K) : V {
        val i = keys.findGE(k)
        if (i < size()) {
            return value(i)
        } else {
            return defaultValue;
        }
    }

    fun getLGE(k: K) : V {
        val i = keys.findLE(k)
        if (i >= 0) {
            return value(i)
        } else {
            return getGE(k);
        }
    }

    override fun toString(): String {
        return keys.toArray()
                .mapIndexed {i,k -> "$k -> ${value(i)}"}
                .joinToString(", ", "[", "]")
    }

    fun copyOfKeys(): AscSortedAppendableArray<K> {
        return keys.copyOf()
    }

    fun copyOfValues(): AppendableArray<V> {
        return values.copyOf()
    }
}

class LLSeries(
        defaultValue: Long = 0L,
        initSize: Int = 1024,
        keys: AscSortedAppendableArray<Long> = LongAscSortedAppendableArray(initSize = initSize),
        values: AppendableArray<Long> = LongAppendableArray(initSize = initSize))
    : Series<Long,Long>(defaultValue, keys, values) {

    constructor(keys: LongArray, values: LongArray)
            : this( keys = LongAscSortedAppendableArray(initValues = keys),
            values = LongAppendableArray(initValues = values))

    fun keysAsLongArray() = (keys as LongAppendableArray).toPrimitiveArray()

    fun valuesAsFloatArray() = (values as LongAppendableArray).toPrimitiveArray()

    companion object Companion {
        fun create(vararg pairs: Pair<Long,Long>): LLSeries {
            val s = LLSeries()
            pairs.forEach { s.add(it.first, it.second) }
            return s
        }
    }
}

class LFSeries(
        defaultValue: Float = 0f,
        initSize: Int = 1024,
        keys: AscSortedAppendableArray<Long> = LongAscSortedAppendableArray(initSize = initSize),
        values: AppendableArray<Float> = FloatAppendableArray(initSize = initSize))
    : Series<Long,Float>(defaultValue, keys, values) {

    constructor(keys: LongArray, values: FloatArray)
            : this( keys = LongAscSortedAppendableArray(initValues = keys),
                    values = FloatAppendableArray(initValues = values))

    fun keysAsLongArray() = (keys as LongAppendableArray).toPrimitiveArray()

    fun valuesAsFloatArray() = (values as FloatAppendableArray).toPrimitiveArray()

    companion object Companion {
        fun create(vararg pairs: Pair<Long,Float>): LFSeries {
            val s = LFSeries()
            pairs.forEach { s.add(it.first, it.second) }
            return s
        }
    }
}
