package com.urbandroid.sleep.sensor.respiration.v2

import com.urbandroid.common.logging.Logger
import com.urbandroid.sleep.sensor.respiration.RespiratoryDetector
import com.urbandroid.sleep.snoring.feature.Moving
import com.urbandroid.sleep.snoring.newfftresult.FftResult
import com.urbandroid.sleep.snoring.newfftresult.FftResultRaw
import com.urbandroid.util.ScienceUtil

import org.jtransforms.fft.FloatFFT_1D

import java.util.Arrays
import kotlin.math.abs
import kotlin.math.max
import kotlin.math.min
import kotlin.math.round

class RespiratoryDetectorV21(
        private val sampleRate: Float,
        private val clock: Clock,
        private val breathLog: BreathLogger?,
        private val listener: RespiratoryDetector.RespiratoryListener?)

    : RespiratoryDetector {

    private val RESP_RATE_FROM = 8
    private val RESP_RATE_TO = 20

    private var firstCall = true
    private var expectedDataSize: Int = 0

    private val BREATH_HISTORY = 10
    private val BREATH_QUORUM_1 = 6
    private val BREATH_QUORUM_2 = 6
    private val APNEA_HISTORY = 20
    private val APNEA_BREATH_QUORUM = 15
    private val MAX_HISTORY = max(BREATH_HISTORY, APNEA_HISTORY)

    private val history = mutableListOf<BreathEvent>()
    private val allHistory = mutableListOf<BreathEvent>()

    override fun detect(data: FloatArray, activity: Int) {
        val t0 = System.currentTimeMillis()
        if (firstCall) {
            processFirstData(data, activity)
        } else {
            processNextData(data, activity)
        }
        val t1 = System.currentTimeMillis()
        //Logger.logInfo("RespiratoryDetectorV21: elapsed time: ${t1-t0}")
    }

    private fun processFirstData(data: FloatArray, activity: Int) {
        if (data.size < 240) {
            Logger.logWarning("RespiratoryDetectorV21: weird small first data. Skipping. " + data.size + " " + activity)
            return
        }
        expectedDataSize = data.size
        firstCall = false
        doProcess(data, activity)
    }

    private fun processNextData(data: FloatArray, activity: Int) {
        if (data.size != expectedDataSize) {
            Logger.logWarning("RespiratoryDetectorV21: unexpected data size: " + data.size + " " + expectedDataSize)
            return
        }
        doProcess(data, activity)
    }

    val avgSignalToNoiseRatio = Moving.avg(10)
    var currentAvgSignalToNoiseRatio = 0f

    val medianBuffer = FloatArrayBuffer(10000)
    val thresholdBuffer = FloatArrayBuffer(10000)

    private fun doProcess(data: FloatArray, activity: Int) {

        val peaklessData = removePeaks(data)

        val fft = fft(peaklessData)
        breathLog?.log(fft.energies.joinToString(",")+"\n")

        val breathFreqFrom = RESP_RATE_FROM.toDouble() / 60
        val breathFreqTo = RESP_RATE_TO.toDouble() / 60
        val contextBinFrom = breathFreqFrom * 0.5
        val contextBinTo = breathFreqTo * 4

        val maxBreathBin = fft.getMaxEnergyBin(contextBinFrom, contextBinTo)
        val maxBreathFreq = fft.getFrequency(maxBreathBin)
        val respRate2x = Math.round(maxBreathFreq * 60).toInt()
        val respRate = if (respRate2x > RESP_RATE_TO) respRate2x / 2 else respRate2x
        val isValidRespRate = respRate in RESP_RATE_FROM .. RESP_RATE_TO

        val event = BreathEvent(clock.getTime(), if (isValidRespRate) respRate else 0)
        history.add(event)
        if (breathLog != null) {
            allHistory.add(event)
        }
        while (history.size > MAX_HISTORY) {
            history.removeAt(0)
        }

        detectHighActivity(event, data)

        detectBreath()

        val contextBinCount = fft.getBinByFrequency(contextBinTo) - fft.getBinByFrequency(contextBinFrom) + 1
        val avgContextEnergy = fft.getEnergySum(contextBinFrom, contextBinTo) / contextBinCount
        val signalToNoiseRatio = fft.getEnergy(maxBreathBin) / avgContextEnergy;
        val validSignalToNoiseRatio = if (event.resolvedAsBreath) signalToNoiseRatio.toFloat() else 2f
        currentAvgSignalToNoiseRatio = avgSignalToNoiseRatio.apply(validSignalToNoiseRatio)

        detectApnea()
    }

    //Remove any peaks above the dynamic threshold. High peaks in the data are caused
    //by intense body movement rather than by breath. They generate a lot of artifacts
    //in the spectrum, which damage the breath signal.
    //FFT on peakless data gives much better results.
    private fun removePeaks(data: FloatArray): FloatArray {
        medianBuffer.add(data)
        val median = ScienceUtil.percentile(medianBuffer.get(), 50f)
        val centeredData = data.map { median - it }.toFloatArray()
        thresholdBuffer.add(centeredData)
        val threshold = max(1f, ScienceUtil.percentile(thresholdBuffer.get(), 50f))
        val peaklessData = data.map { min(max(it - median, -threshold), threshold) }.toFloatArray()
        return peaklessData
    }

    private val lowActivity = Moving.min(5)

    //Very naive algorithm, but I did not find anything better, as the variance of the energy is highly irregular.
    private fun detectHighActivity(event: BreathEvent, data: FloatArray) {
        val median = ScienceUtil.percentile(data, 50f)
        val activity = data.map { abs(it - median) }.sum()
        val lowActivity = lowActivity.apply(activity)
        val threshold = lowActivity * 1.75f
        event.isHighActivity = activity > threshold
    }

    //Breath detection algorithm:
    //
    // Each BreathEvent corresponds to a 1 minute interval and they overlap by 30 sec.
    // We analyze the last BREATH_HISTORY intervals in each step.
    //
    // An interval is marked as breath iff:
    // * The maximum frequency on the FFT falls into the expected breath rate (8-20) or its double.
    // * There are enough intervals (BREATH_QUORUM_2) with a similar breath rate (+-1) in the recent history.
    private fun detectBreath() {

        val history = history.slice(max(0, history.size-BREATH_HISTORY) until history.size)

        val validRates =  history
                .filter { it.isValidRespRate() }
                .map { it.respRate.toFloat() }
                .toTypedArray()

        if (validRates.size < BREATH_QUORUM_1) {
            return
        }

        val medianRate = round(ScienceUtil.percentile(validRates, 50f)).toInt()
        val realBreathFrom = medianRate - 1
        val realBreathTo = medianRate + 1

        val realBreathCandidates = history
                .filter { it.respRate in realBreathFrom .. realBreathTo }

        if (realBreathCandidates.size < BREATH_QUORUM_2) {
            return
        }

        realBreathCandidates
                .filter { !it.resolvedAsBreath }
                .forEach {
                    it.resolvedAsBreath = true
                    listener?.onBreathDetected(it.timestamp, it.respRate)
                }
    }

    //Apnea detection algorithm:
    //
    //An interval is marked as apnea iff:
    // * The last event is breath
    // * The previous event is not breath, and not a high activity either
    // * Most of the recent records were breaths (i.e. there is a good quality signal)
    private fun detectApnea() {

        if (history.size < APNEA_HISTORY) {
            return
        }
        val history = history.slice(max(0, history.size-APNEA_HISTORY) until history.size)

        val apneaCandidate =
                history.get(history.size-1).resolvedAsBreath &&
                ! history.get(history.size-2).resolvedAsBreath &&
                ! history.get(history.size-2).isHighActivity &&
                history.filter { it.resolvedAsBreath }.size > APNEA_BREATH_QUORUM

        if (apneaCandidate) {
            Logger.logInfo("RespiratoryDetectorV21 apnea candidate, signalStrength=$currentAvgSignalToNoiseRatio")
            //Additional criterion - there has been relatively clear breath signal recently
            //(high signal/noise ratio), so the outage is more likely to be a breath disturbance,
            //rather than just random noise.
            if (currentAvgSignalToNoiseRatio > 4.0) {
                val apneaEvent = history.get(history.size - 2)
                apneaEvent.resolvedAsApnea = true
                listener?.onApneaDetected(apneaEvent.timestamp, 1)
            }
        }
    }

    private fun fft(data: FloatArray): FftResult {
        val dataFft = Arrays.copyOf(data, data.size)
        val dataWindowFft = FloatFFT_1D(dataFft.size.toLong())
        dataWindowFft.realForward(dataFft)
        val rawResult = FftResultRaw(dataFft, sampleRate)
        return rawResult
    }

    override fun dataBroken() {
        history.clear()
    }

    override fun trackingFinished() {
        logDetailedResults()
    }

    private fun logDetailedResults() {

        if (breathLog != null) {

            val breathAll = allHistory.map { it.respRate }
            breathLog.log(breathAll.joinToString(",") + "\n")

            val breathResolved = allHistory.map { if (it.resolvedAsBreath) it.respRate else 0 }
            breathLog.log(breathResolved.joinToString(",") + "\n")

            val apneas = allHistory.map { if (it.resolvedAsApnea) 1 else 0 }
            breathLog.log(apneas.joinToString(",") + "\n")

            val highActivities = allHistory.map { if (it.isHighActivity) 1 else 0 }
            breathLog.log(highActivities.joinToString(",") + "\n")
        }
    }

    private class BreathEvent(val timestamp: Long, val respRate:Int) {

        var isHighActivity = false
        var resolvedAsBreath = false
        var resolvedAsApnea = false

        fun isValidRespRate():Boolean {
            return respRate > 0
        }
    }
}
