/*
 * Copyright (C) 2013 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.urbandroid.sleep.service

import android.content.BroadcastReceiver
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.os.Build
import android.os.PowerManager
import android.util.Log
import android.util.SparseArray

abstract class WakefulForegroundBroadcastReceiver : BroadcastReceiver() {
    companion object {
        private const val EXTRA_WAKE_LOCK_ID = "android.support.content.wakelockid"

        private val sActiveWakeLocks = SparseArray<PowerManager.WakeLock>()
        private var mNextId = 1

        fun startWakefulService(context: Context, intent: Intent): ComponentName? {
            synchronized(sActiveWakeLocks) {
                val id = mNextId
                mNextId++
                if (mNextId <= 0) {
                    mNextId = 1                                  
                }

                intent.putExtra(EXTRA_WAKE_LOCK_ID, id)

                val comp: ComponentName = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                    context.startForegroundService(intent) ?: return null
                } else {
                    context.startService(intent) ?: return null
                }

                val pm = context.getSystemService(Context.POWER_SERVICE) as PowerManager
                val wl = pm.newWakeLock(PowerManager.PARTIAL_WAKE_LOCK,
                        "wake:" + comp.flattenToShortString())
                wl.setReferenceCounted(false)
                wl.acquire((60 * 1000).toLong())
                sActiveWakeLocks.put(id, wl)
                return comp
            }
        }

        fun completeWakefulIntent(intent: Intent): Boolean {
            val id = intent.getIntExtra(EXTRA_WAKE_LOCK_ID, 0)
            if (id == 0) {
                return false
            }
            synchronized(sActiveWakeLocks) {
                val wl = sActiveWakeLocks.get(id)
                if (wl != null) {
                    wl.release()
                    sActiveWakeLocks.remove(id)
                    return true
                }
                // We return true whether or not we actually found the wake lock
                // the return code is defined to indicate whether the Intent contained
                // an identifier for a wake lock that it was supposed to match.
                // We just log a warning here if there is no wake lock found, which could
                // happen for example if this function is called twice on the same
                // intent or the process is killed and restarted before processing the intent.
                Log.w("WakefulBroadcastReceiv.", "No active wake lock id #" + id)
                return true
            }
        }
    }
}
