package com.urbandroid.util

import com.urbandroid.common.logging.Logger

import java.util.concurrent.atomic.AtomicLong

/**
 * A utility to prevent flooding the logs with a lot of same messages.
 * If the logging methods are called repeatedly, most of the calls get ignored,
 * and the ratio of the messages that actually get logged gradually decreases.
 *
 * TODO Actually, it might be better to implement this logic directly into our Logger.
 * For example, to register there some kind of stateful filters (a regex and a counter)
 * that would match the messages and discard them if there are too many of them.
 */
class ConsiderateLogger @JvmOverloads constructor(private val maxPrune: Long = 1024) {

    private val pruneFactor: AtomicLong
    private val count: AtomicLong

    init {
        pruneFactor = AtomicLong(1)
        count = AtomicLong(0)
    }

    fun reset() {
        pruneFactor.set(1)
        count.set(0)
    }

    fun info(s: String) {
        pruneAndRun {Logger.logInfo(s)}
    }

    fun severe(e: Throwable) {
        pruneAndRun {Logger.logSevere(e)}
    }

    private fun pruneAndRun(command: () -> Unit) {
        if (count.incrementAndGet() % pruneFactor.get() == 0L) {
            command()
            if (pruneFactor.get() < maxPrune) {
                pruneFactor.set(pruneFactor.get() * 2)
            }
        }
    }
}
